# -*- coding: utf-8 -*-
# Copyright (c) 2021 Marek Wadinger

# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from typing import Union

import numpy as np
import pandas as pd

import plotly.express as px
import plotly

from pandas.core.frame import DataFrame

from sklearn.model_selection import train_test_split as tt_split
from sklearn.preprocessing import MinMaxScaler

from pyod.models.copod import COPOD
from pyod.models.cblof import CBLOF
from pyod.models.hbos import HBOS
from pyod.models.iforest import IForest
from pyod.models.knn import KNN
from pyod.models.pca import PCA
from pyod.models.ocsvm import OCSVM
from pyod.models.rod import ROD

from src.seanomalydetection.se_api_process import get_se_as_df
from src.seanomalydetection.anomaly_plot import plot_outlier_detection_plotly

plotly.offline.init_notebook_mode(connected=True)


class Report:
    pass


def list_na_columns(df: pd.DataFrame, max_na_fraction: float = 0.1)\
        -> np.ndarray:
    """List invalid columns due to number of NAs over given threshold.

    Parameters
        ----------
        df:
            pd.DataFrame
        max_na_fraction:
            float maximum fraction of NAs in column to be considered valid.

    Returns
        -------
        np.ndarray listing names of invalid columns.
    """
    if max_na_fraction >= 1:
        max_na_fraction = max_na_fraction / 100

    sum_na_cols = df.isna().sum()

    return sum_na_cols[sum_na_cols / df.shape[0] > max_na_fraction].index.values


def dropna_columns(df: pd.DataFrame, max_na_fraction: float = 0.1)\
        -> pd.DataFrame:
    """Drop invalid columns due to number of NAs over given threshold.

    Parameters
        ----------
        df:
            pd.DataFrame
        max_na_fraction:
            float maximum fraction of NAs in column to be considered valid.

    Returns
        -------
        pd.DataFrame without invalid columns.
    """
    cols_to_drop = list_na_columns(df, max_na_fraction)

    return df.drop(columns=cols_to_drop)


def drop_missing(df: pd.DataFrame, max_na_fraction: float = 0.1)\
        -> pd.DataFrame:
    """Drop invalid columns given max number of NAs allowed and rows with NAs.

    Parameters
        ----------
        df:
            pd.DataFrame
        max_na_fraction:
            float maximum fraction of NAs in column to be considered valid.

    Returns
        -------
        pd.DataFrame with dropped invalid columns and NA entries.
    """
    df = dropna_columns(df, max_na_fraction=max_na_fraction)

    return df.dropna()


def drop_extreme_outs(df: pd.DataFrame):
    df_strings = filter_cols_by_name(df, "String")
    df_diffs = np.divide(df_strings.sum(axis=1), df["Inverter 1"]).to_frame("yields")
    df = df[df_diffs['yields'].between((df_diffs.mean()
                                        - 3 * df_diffs.std())[0],
                                       (df_diffs.mean()
                                        + 3 * df_diffs.std())[0])]
    return df


def filter_cols_by_name(df: pd.DataFrame, name: str) -> pd.DataFrame:
    """Return df with columns containing name.

    Parameters
        ----------
        df:
            pd.DataFrame with columns containing name.
        name:
            string, name specific for columns to filter.

    Returns
        -------
        pd.DataFrame with product of row-wise division
            with the same shape as df_subsets.
    """
    # old behavior
    # return df[[i for i in df if name in i]]
    return df.loc[:, df.columns.str.contains(name)]


def get_yield(df: pd.DataFrame, name_subset: str, name_set: str)\
        -> pd.DataFrame:
    """Return product of row-wise division.

    Perform row-wise division between subsets and corresponding set.

    Parameters
        ----------
        df:
            pd.DataFrame with column names containing: ".*[name_subset]\..*"
            and ".*[name_set]\..*".
        name_subset:
            string, name specific for columns of subsets.
        name_set:
            string, name specific for columns of sets.

    Returns
        -------
        pd.DataFrame with product of row-wise division
            with the same shape as df_subsets.
    """

    df_subsets = filter_cols_by_name(df, name_subset)
    df_sets = filter_cols_by_name(df, name_set)

    subsets_yield = []
    # Works with arrays for speed
    for set_id in df_sets:
        set_num = set_id.split(" ")[1] + "."
        # Returns array of modules belonging to given string
        subsets = filter_cols_by_name(df_subsets, set_num).T.values
        subsets_yield = [*subsets_yield,
                         *np.divide(subsets, df_sets[set_id].values)
                         * 5 * subsets.shape[0]]

    return pd.DataFrame(np.transpose(subsets_yield),
                        index=df_subsets.index, 
                        columns=df_subsets.columns.str.replace("Module", "Yield"))


def json_to_df(
        data: Union[DataFrame, str] = None,
        max_na_fraction: float = 0.1,
        name_subset: str = "Module",
        name_set: str = "String"
) -> pd.DataFrame:
    """Return dfs containing modules and yields.

        Performs preprocessing, feature engineering on input data.

        Parameters
            ----------
            data:
                string containing relative or absolute path to json file
                or pd.DataFrame.
            max_na_fraction:
                float maximum fraction of NAs in column to be considered valid.
            name_subset:
                string, name specific for columns of subsets.
            name_set:
                string, name specific for columns of sets.

        Returns
            -------
            df_modules:
                pd.DataFrame with subset of input df containing only modules.
            df_yields:
                pd.DataFrame with individual modules contribution toward
                whole production on string.
        """
    if isinstance(data, DataFrame):
        df = data
    elif isinstance(data, str):
        df = get_se_as_df(data)
    else:
        raise TypeError("Wrong data type.")

    df = drop_missing(df, max_na_fraction)

    df = drop_extreme_outs(df)

    df_modules = filter_cols_by_name(df, name_subset)

    df_yields = get_yield(df, name_subset, name_set)
    # df_yields.columns = df_yields.columns.str.replace("Module", "Yield")
    df = pd.concat([df_modules, df_yields], axis=1)

    return df


def train_test_split(
        df_floats: pd.DataFrame,
        train_size: float = 0.8,
        random_state_num: int = 42,
        shuffle: bool = True,
):
    """Performs train, test split on dataset.

        Parameters
            ----------
            df_floats:
                pd.DataFrame with floats as the elements.
            train_size:
                float, fraction of dataset to be used for training
                anomaly detection model.
            random_state_num:
                int, random seed used to initialize generator.
            shuffle:
                shuffle the data before splitting.

        Returns
            -------
            pd.DataFrames.
        """
    if train_size >= 1:
        train_size = train_size / 100
    if random_state_num:
        random_state = np.random.RandomState(random_state_num)
    else:
        random_state = None

    df_train, df_test = tt_split(df_floats,
                                 train_size=train_size,
                                 random_state=random_state,
                                 shuffle=shuffle)
    return df_train, df_test


def pick_model(
        classifier: str = "K Nearest Neighbors",
        outliers_rate: float = 0.0046,
        random_state_num: int = 42
):
    """Return classifier object.

    Parameters
        ----------
        classifier:
            string representing name of anomaly detection algorithm.
        outliers_rate:
            proportion of training set to be considered outlier.
        random_state_num:
            int, random seed used to initialize generator.

    Returns
        -------
        classifier object for outlier detection.
    """
    if outliers_rate >= 1:
        outliers_rate = outliers_rate / 100

    if random_state_num:
        random_state = np.random.RandomState(random_state_num)
    else:
        random_state = None
    
    classifiers = {
        "Average KNN (AKNN)":
            KNN(method="mean", contamination=outliers_rate),
        "Cluster-based Local Outlier Factor (CBLOF)":
            CBLOF(contamination=outliers_rate, check_estimator=False,
                  random_state=random_state),
        "Copula based Outlier Detection (COPOD)":
            COPOD(contamination=outliers_rate),
        "Histogram-base Outlier Detection (HBOS)":
            HBOS(contamination=outliers_rate),
        "Isolation Forest (IForest)":
            IForest(contamination=outliers_rate, random_state=random_state),
        "K Nearest Neighbors (KNN)":
            KNN(contamination=outliers_rate),
        "One-Class SVM (OCSVM)":
            OCSVM(contamination=outliers_rate),
        "Principal component analysis (PCA)":
            PCA(contamination=outliers_rate),
        "Rotation-based Outlier Detector (ROD)":
            ROD(contamination=outliers_rate),
    }
    
    if classifier != "all":
        clf_name = ""
        for name in classifiers.keys():
            if classifier in name:
                clf_name = name
                break
        if clf_name:
            clf = classifiers.get(clf_name)

        else:
            raise NameError(
                "Unknown classifier. "
                "Please use one of those: {}.".format(list(classifiers.keys()))
            )

    else:
        raise Warning(
            "This option is currently unsupported."
            "\nPlease use one of those classifiers:"
            "\n{}.".format(list(classifiers.keys()))
        )
        # for i, (clf_name, clf) in enumerate(classifiers.items()):
        #    # fit model
        #    clf.fit(df_train)
        #    # prediction of a datapoint category outlier or inlier
        #    y_labels = clf.predict(df_scaled)
        #    plot_outlier_detection(df_scaled, y_labels, clf, clf_name, scaler)
    return clf


def fit_model(
        df_floats: pd.DataFrame,
        classifier: str = "K Nearest Neighbors",
        outliers_rate: float = 0.0046,
        random_state_num: int = 42
):
    """Return fitted classifier object.

    Parameters
        ----------
        df_floats:
            pd.DataFrame with floats as the elements.
        classifier:
            string representing name of anomaly detection algorithm.
        outliers_rate:
            proportion of training set to be considered outlier.
        random_state_num:
            int, random seed used to initialize generator.

    Returns
        -------
        fitted classifier object for outlier detection.
    """
    if df_floats.shape[0] < 10:
        raise Warning("Not enough measurements. "
                      "Please use DataFrame with at last 10 measurements.")

    clf = pick_model(classifier, outliers_rate, random_state_num)
    clf.fit(df_floats)

    return clf


def estimate_outliers_rate(
        df_floats,
        classifier: str = "K Nearest Neighbors",
        outliers_rate: float = 0.0046,
        random_state_num: int = 42,
        plot_est_heatmap: bool = False
) -> pd.Series:
    """Return estimated outliers rates for all entities in dataframe.

    Parameters
        ----------
        df_floats:
            pd.DataFrame with floats as the elements.
        classifier:
            string representing name of anomaly detection algorithm.
        outliers_rate:
            proportion of training set to be considered outlier.
        random_state_num:
            int, random seed used to initialize generator.
        plot_est_heatmap:
            plots 2d contourf of estimators detection scores.
    Returns
        -------
        pd.Series containing estimated outliers rates assigned to entities.
    """
    df_modules = filter_cols_by_name(df_floats, "Module")
    df_yields = filter_cols_by_name(df_floats, "Yield")
    df_modules_l = pd.melt(df_modules.reset_index(), id_vars="created_on")
    df_yields_l = pd.melt(df_yields.reset_index(), id_vars="created_on")
    df_modules_l = df_modules_l.assign(yields=df_yields_l["value"])

    clf = fit_model(df_modules_l[["value", "yields"]],
                    classifier, outliers_rate, random_state_num)

    y_labels = clf.predict(df_modules_l[["value", "yields"]])
    y_scores = clf.decision_function(df_modules_l[["value", "yields"]])

    if plot_est_heatmap:
        plot_outlier_detection_plotly(df_modules_l[["value", "yields"]],
                                      y_labels,
                                      clf,
                                      classifier,
                                      None,
                                      df_modules_l)

    df_modules_l = df_modules_l.assign(outlier_score=y_scores)

    df_scores_temp = df_modules_l.pivot("created_on", "variable", "outlier_score")

    outlier_rates = pd.Series(index=df_scores_temp.columns)

    for i in df_scores_temp.columns.values:
        mean_score = df_scores_temp[i].mean()
        std_score_all = df_scores_temp.std().mean()
        outs = df_scores_temp[i][df_scores_temp[i]
                                 < mean_score - 3 * std_score_all].count()
        outs = outs + df_scores_temp[i][df_scores_temp[i]
                                        > mean_score + 3 * std_score_all].count()
        outs = outs / df_scores_temp.__len__()

        # Upper and lower bound handling.
        if outs > 0.5:
            outs = 0.5
        elif outs == 0:
            outs = 0.00001

        outlier_rates[i] = outs

    return outlier_rates


def train_model(
        df_train: pd.DataFrame,
        classifier: str = "K Nearest Neighbors",
        outliers_rate: float = 0.0046,
        random_state_num: int = 42,
        plot_est_heatmap: bool = False
) -> pd.Series:
    """Return trained ensemble of classifiers.

    Parameters
        ----------
        df_train:
            pd.DataFrame with floats as the elements.
        classifier:
            string representing name of anomaly detection algorithm.
        outliers_rate:
            proportion of training set to be considered outlier.
        random_state_num:
            int, random seed used to initialize generator.
        plot_est_heatmap:
            plots 2d contourf of estimators detection scores.

    Returns
        -------
        pd.Series containing colassifiers assigned to entity.
    """

    outs = estimate_outliers_rate(df_train,
                                  classifier,
                                  outliers_rate,
                                  random_state_num,
                                  plot_est_heatmap)

    models = pd.Series(index=outs.index)

    for i in outs.index.values:
        id_num = i.replace("Module ", "")
        clf = fit_model(df_train[[col for col in df_train if col.endswith(id_num)]],
                        classifier, outs[i], random_state_num)

        models[i] = clf

    return models


def evaluate_models(
        df_test: pd.DataFrame,
        models):
    """Return dfs containing outlier tags and outlier scores.

    Parameters
        ----------
        df_test:
            pd.DataFrame.
        models:
            pd.Series containing ensemble of models for specified entities
    Returns
        -------
        df_outliers:
            pd.DataFrame containing outlier labels.
        df_scores:
            pd.DataFrame containing outlier scores.
    """
    list_columns = [col for col in df_test.columns.values if "Module" in col]

    df_outliers = pd.DataFrame(index=df_test.index, columns=list_columns)
    df_scores = pd.DataFrame(index=df_test.index, columns=list_columns)

    for i in list_columns:
        id_num = i.replace("Module ", "")

        y_labels = models[i].predict(
            df_test.loc[:, df_test.columns.str.endswith(id_num)])

        y_scores = models[i].decision_function(
            df_test.loc[:, df_test.columns.str.endswith(id_num)])

        df_outliers[i] = y_labels
        df_scores[i] = y_scores

    return df_outliers, df_scores


def report_anomaly(df_outliers: pd.DataFrame) -> Report:
    """Return report object and prints out reporting.

    Parameters
        ----------
        df_outliers:
            pd.DataFrame in wide format containing outlier labels.

    Returns
        -------
        Report object with reporting entities.
    """
    if df_outliers.astype(bool).sum().sum() > 0:
        report_obj = Report()
        report_obj.anomaly_rate = round((df_outliers.values == 1).sum()
                                        / df_outliers.size * 100, 2)

        report_obj.bad_module = df_outliers.astype(bool).sum().idxmax()
        sum_bad_module_anomaly = df_outliers[report_obj.bad_module].sum()
        sum_bad_module = df_outliers[report_obj.bad_module].count()
        report_obj.bad_module_rate = round(sum_bad_module_anomaly
                                           / sum_bad_module * 100, 2)

        report_obj.bad_day = df_outliers.astype(bool).sum(axis=1).idxmax()
        sum_bad_day_anomaly = df_outliers.loc[report_obj.bad_day].sum()
        sum_bad_day = df_outliers.loc[report_obj.bad_day].count()
        report_obj.bad_day_rate = round(sum_bad_day_anomaly
                                        / sum_bad_day * 100, 2)

        print(
            """\nREPORT\
            \n--------------------
            \nAnomaly rate of all modules in dataset is {:.2f}%.\
            \n\n{} requires your attention!\
            \n{:.2f}% of observations of module were considered outliers.\
            \n\n{} is the most suspicious day.\
            \n{:.2f}% of observations of that day were considered outliers.\
            \n\nCheck out the returned dataframe with anomalous observations.\
            \n--------------------\
            """.format(
                report_obj.anomaly_rate,
                report_obj.bad_module,
                report_obj.bad_module_rate,
                report_obj.bad_day.strftime("%Y-%m-%d"),
                report_obj.bad_day_rate,
            )
        )

    else:
        report_obj = Report()
        print(
            """\nREPORT\
                    \n--------------------\
                    \nCongrats, your modules work just fine. :)"""
        )

    return report_obj


def list_most_anomalous(df_outliers: pd.DataFrame, min_anomalies: int = 5)\
        -> list:
    df_anomaly = df_outliers.sum()[df_outliers.sum() > 0]
    most_anomalous = []
    for i in df_anomaly.index.values:
        if df_anomaly[i] > min_anomalies:
            most_anomalous.append(i)

    return most_anomalous


def get_sma(
        df_scores: Union[pd.DataFrame, pd.Series],
        window: int = 7,
):

    return df_scores.rolling(window=window).mean()


def scale_df(df_floats: pd.DataFrame, scale: str = None):
    """Return scaled pd.DataFrame.

    Parameters
        ----------
        df_floats:
            pd.DataFrame with floats as the elements.
        scale:
            str defining scaling option.

    Returns
        -------
        scaled pd.DataFrame.
        scaler object.
    """
    if scale == "max":
        df_scaled = df_floats/df_floats.max()
        scaler = None
    elif scale == "min":
        df_scaled = df_floats/df_floats.min()
        scaler = None
    elif scale == "minmax":
        scaler = MinMaxScaler(feature_range=(0, 1))
        scaled = scaler.fit_transform(df_floats)
        df_scaled = pd.DataFrame(scaled,
                                 index=df_floats.index, columns=df_floats.columns)
    else:
        df_scaled = df_floats
        scaler = None

    return df_scaled, scaler


def plot_anomaly(
        df_outliers: pd.DataFrame,
        df_scores: pd.DataFrame,
        window: int = 7,
        min_anomalies: int = 5,
        scale: Union[str, None] = None
):
    """Return timeline plot containing modules with at last 5 anomalous
    observations.

    Parameters
        ----------
        df_outliers:
            pd.DataFrame in wide format containing outlier labels.
        df_scores:
            pd.DataFrame in wide format containing outlier scores.
        window:
            int length of moving window
        min_anomalies:
            int minimum number of anomalies to printout.
        scale:
            str specifying scaling option.
    """
    most_anomalous = list_most_anomalous(df_outliers, min_anomalies)

    df_sma_scores = get_sma(df_scores[most_anomalous], window=window)

    df_sma_scores, scaler = scale_df(df_sma_scores, scale)

    if not df_sma_scores.empty:
        fig = px.line(
            pd.melt(df_sma_scores.reset_index(), id_vars="created_on",
                    var_name="Module ID", value_name="Outlier Score").dropna(),
            x="created_on",
            y="Outlier Score",
            color="Module ID",
            title=f"Daily outlier score "
                  f"({window}-day moving average) of anomalous modules",
            labels={"created_on": "Date of observation"}
        )
        fig2 = px.line(
            pd.melt(df_scores[df_sma_scores.columns].reset_index(), id_vars="created_on",
                    var_name="Module ID", value_name="Outlier Score").dropna(),
            x="created_on",
            y="Outlier Score",
            color="Module ID",
            title=f"Daily outlier score "
                  f"({window}-day moving average) of anomalous modules",
            labels={"created_on": "Date of observation"}
        )
        fig2.update_traces(dict(line_width=0.4))
        for i in range(0, len(fig2.data)):
            fig.add_trace(fig2.data[i])
        plotly.offline.plot(fig)
    return


def tidy_detection(
        data: Union[DataFrame, str] = None,
        max_na_fraction: float = 0.1,
        name_subset: str = "Module",
        name_set: str = "String",
        train_size: float = 0.8,
        random_state_num: int = 42,
        shuffle: bool = True,
        classifier: str = "K Nearest Neighbors",
        outliers_rate: float = 0.0046,
        plot_est_heatmap: bool = False,
        report: bool = False,
        plot: bool = False,
        window: int = 7,
        min_anomalies: int = 5,
        scale: Union[str, None] = None
):
    """Return dfs containing modules and yields.

    Performs preprocessing, feature engineering on input data.

    Parameters
        ----------
        data:
            string containing relative or absolute path to json file
            or pd.DataFrame.
        max_na_fraction:
            float maximum fraction of NAs in column to be considered valid.
        name_subset:
            string, name specific for columns of subsets.
        name_set:
            string, name specific for columns of sets.
        train_size:
            float, fraction of dataset to be used for training
            anomaly detection model.
        random_state_num:
            int, random seed used to initialize generator.
        shuffle:
            shuffle the data before splitting.
        classifier:
            string representing name of anomaly detection algorithm.
        outliers_rate:
            proportion of training set to be considered outlier.
        plot_est_heatmap:
            plots 2d contourf of estimators detection scores.
        report:
            bool that specifies whether or not to print out report.
        plot:
            plots 2d contourf of anomaly detection scores.
        window:
            int length of moving window
        min_anomalies:
            int minimum number of anomalies to printout.
        scale:
            str specifying scaling option.
    Returns
        -------
        df_outliers:
            pd.DataFrame containing outlier labels.
        df_scores:
            pd.DataFrame containing outlier scores.
    """
    df = json_to_df(data, max_na_fraction, name_subset, name_set)

    df_train, df_test = train_test_split(df, train_size,
                                         random_state_num, shuffle)

    models = train_model(df_train, classifier, outliers_rate, random_state_num,
                         plot_est_heatmap)

    df_outliers, df_scores = evaluate_models(df, models)

    if report:
        report_anomaly(df_outliers)

    if plot:
        plot_anomaly(df_outliers, df_scores, window, min_anomalies, scale)

    return df_outliers, df_scores


def detect_anomaly(
        df_floats: pd.DataFrame,
        train_size: float = 0.8,
        random_state_num: int = 42,
        shuffle: bool = True,
        outliers_rate: float = 0.0046,
        classifier: str = "K Nearest Neighbors",
        plot_heatmap: bool = False,
        df: pd.DataFrame = None
):
    """Return binary classified outlier and raw outlier score.

    Performs training of anomaly detection model on subset of dataset
    and returns binary label and decision score for whole dataset.

    Parameters
        ----------
        df_floats:
            pd.DataFrame with floats as the elements.
        train_size:
            float, fraction of dataset to be used for training
            anomaly detection model.
        random_state_num:
            int, random seed used to initialize generator.
        shuffle:
            shuffle the data before splitting.
        outliers_rate:
            proportion of training set to be considered outlier.
        classifier:
            string representing name of anomaly detection algorithm.
        plot_heatmap:
            plots 2d contourf of anomaly detection scores.
        df:
            pd.DataFrame from first argument with all columns
    Returns
        -------
        y_labels:
            numpy array of the same length as df_floats that assigns
            0/1 (inlier/outlier) to each observation according to fitted model.
        y_scores:
            numpy array of the same length as df_floats that assigns
            outlier scores to each observation according to fitted model.
    """
    # df_scaled, scaler = scale_df(df_floats, "minmax")

    df_train, df_test = train_test_split(df_floats,
                                         train_size=train_size,
                                         random_state_num=random_state_num,
                                         shuffle=shuffle)

    clf = fit_model(df_train, classifier, outliers_rate, random_state_num)

    y_labels = clf.predict(df_floats)
    y_scores = clf.decision_function(df_floats)

    if plot_heatmap:
        plot_outlier_detection_plotly(df_floats, y_labels, clf, None, None, df)

    return y_labels, y_scores


def tidy_detection_old(
        data: Union[DataFrame, str] = None,
        max_na_fraction: float = 0.1,
        name_subset: str = "Module",
        name_set: str = "String",
        train_size: float = 0.8,
        random_state_num: int = 42,
        shuffle: bool = True,
        outliers_rate: float = 0.0046,
        classifier: str = "K Nearest Neighbors",
        report: bool = False,
        plot: bool = False,
):
    """Return dfs containing modules, yields, outlier tags and outlier scores.

    Performs preprocessing, feature engineering on dataset
    and trains anomaly detection model on randomly selected subset of dataset.
    Returns subset of df that is considered anomalous with report.

    Parameters
        ----------
        data:
            pd.DataFrame.
        max_na_fraction:
            float maximum fraction of NAs in column to be considered valid.
        name_subset:
            string, name specific for columns of subsets.
        name_set:
            string, name specific for columns of sets.
        train_size:
            proportion of dataset to be used for training
            anomaly detection model.
        random_state_num:
            int, random seed used to initialize generator.
        shuffle:
            shuffle the data before splitting.
        outliers_rate:
            proportion of training set to be considered outlier.
        classifier:
            string representing name of anomaly detection algorithm.
        report:
            bool that specifies whether or not to print out report.
        plot:
            plots 2d contourf of anomaly detection scores.

    Returns
        -------
        df_modules:
            pd.DataFrame with subset of input df containing only modules.
        df_yields:
            pd.DataFrame with individual modules contribution toward
            whole production on string.
        df_outliers:
            pd.DataFrame containing outlier labels.
        df_scores:
            pd.DataFrame containing outlier scores.
    """
    df = json_to_df(data, max_na_fraction, name_subset, name_set)

    df_modules = filter_cols_by_name(df, "Module")
    df_yields = filter_cols_by_name(df, "Yield")

    df_modules_long = pd.melt(df_modules.reset_index(), id_vars="created_on")
    df_yields_long = pd.melt(df_yields.reset_index(), id_vars="created_on")
    df_modules_long = df_modules_long.assign(yields=df_yields_long["value"])

    y_labels, y_scores = detect_anomaly(
        df_modules_long[["value", "yields"]],
        train_size,
        random_state_num,
        shuffle,
        outliers_rate,
        classifier,
        True,
        df_modules_long
    )

    df_modules_long = df_modules_long.assign(outlier=y_labels)
    df_modules_long = df_modules_long.assign(outlier_score=y_scores)

    df_outliers = df_modules_long.pivot("created_on", "variable", "outlier")
    df_scores = df_modules_long.pivot("created_on", "variable", "outlier_score")

    if report:
        report_anomaly(df_outliers)
    if plot:
        plot_anomaly(df_outliers, df_scores)

    return df_modules, df_yields, df_outliers, df_scores


def tidy_individual_detection_old(
        data: Union[DataFrame, str, None] = None,
        max_na_fraction: float = 0.1,
        name_subset: str = "Module",
        name_set: str = "String",
        train_size: float = 0.8,
        random_state_num: int = 42,
        shuffle: bool = True,
        outliers_rate: float = 0.0046,
        classifier: str = "K Nearest Neighbors",
        report: bool = False,
        plot: bool = False,
):
    """Return df containing outlier scores, outlier tags and report.

    Performs preprocessing, feature engineering on dataset and trains
    anomaly detection model on randomly selected subset of dataset.
    Returns subset of df that was considered anomalous with report.

    Parameters
        ----------
        data:
            pd.DataFrame.
        max_na_fraction:
            float maximum fraction of NAs in column to be considered valid.
        name_subset:
            string, name specific for columns of subsets.
        name_set:
            string, name specific for columns of sets.
        train_size:
            proportion of dataset to be used for training
            anomaly detection model.
        random_state_num:
            int, random seed used to initialize generator.
        shuffle:
            shuffle the data before splitting.
        outliers_rate:
            proportion of training set to be considered outlier.
        classifier:
            string representing name of anomaly detection algorithm.
        report:
            bool that specifies whether or not to print out report.
        plot:
            plots 2d contourf of anomaly detection scores.
    Returns
        -------
        df_modules:
            pd.DataFrame with subset of input df containing only modules.
        df_yields:
            pd.DataFrame with individual modules contribution toward whole
            production on string.
        df_outliers:
            pd.DataFrame containing outlier labels.
        df_scores:
            pd.DataFrame containing outlier scores.
    """
    df_modules, df_yields, \
        df_outliers_temp, df_scores_temp = tidy_detection_old(
            data, max_na_fraction,
            name_subset, name_set,
            train_size, random_state_num, shuffle,
            outliers_rate, classifier,
            False, False
        )

    df_outliers = pd.DataFrame(index=df_modules.index,
                               columns=df_modules.columns)
    df_scores = pd.DataFrame(index=df_modules.index,
                             columns=df_modules.columns)

    for i in df_modules.columns.values:
        mean_score = df_scores_temp[i].mean()
        std_score_all = df_scores_temp.std().mean()
        outs = df_scores_temp[i][df_scores_temp[i]
                                 < mean_score - 3 * std_score_all].count()
        outs = outs + df_scores_temp[i][df_scores_temp[i]
                                        > mean_score + 3 * std_score_all].count()
        outs = outs / df_modules.__len__()

        # Upper and lower bound handling.
        if outs > 0.5:
            outs = 0.5
        elif outs == 0:
            outs = 0.00001

        y_labels, y_scores = detect_anomaly(
            pd.concat([df_modules[i],
                       df_yields[i.replace("Module", "Yield")]],
                      axis=1),
            train_size,
            random_state_num,
            shuffle,
            outs,
            classifier,
            False,
        )

        df_outliers[i] = y_labels
        df_scores[i] = y_scores

    if report:
        report_anomaly(df_outliers)
    if plot:
        plot_anomaly(df_outliers, df_scores)

    return df_modules, df_yields, df_outliers, df_scores


if __name__ == "__main__":
    outliers, scores = tidy_detection("se_daily.json")
    print(scores.head())
