import numpy as np
import pandas as pd

import plotly
import plotly.express as px
import plotly.graph_objects as go
from scipy.stats import stats

import matplotlib
from matplotlib import pyplot as plt

from sklearn.preprocessing import MinMaxScaler


def plot_outlier_detection_plotly(
        df_floats: pd.DataFrame,
        y_pred: np.ndarray,
        clf,
        clf_name: str = None,
        scaler=None,
        df=None
):
    """ Return contourf plot of the anomaly detection model.

    Plots contourf plot of decision scores marking area
    where data points would be considered inliers.

    Parameters
        ----------
        df_floats:
            pd.DataFrame with elements as floats.
        y_pred:
            numpy array of the same length as df_floats that assigns 0/1
            (inlier/outlier) to each observation according to fitted model.
        clf:
            fitted model.
        clf_name:
            name of fitted model.
        scaler:
            estimator that was used to change range of df_floats DataFrame
        df:
            pd.DataFrame from first argument with all columns

    Returns
        -------
        Contourf plot.
    """
    if df_floats.shape[1] > 2:
        print('Plotting first two variables')
    elif df_floats.shape[1] < 2:
        print('Sorry can not plot less than one variable')

    # predict raw anomaly score
    scores_pred = clf.decision_function(df_floats.iloc[:, [0, 1]]) * -1
    # threshold value to consider a datapoint inlier or outlier
    threshold = stats.scoreatpercentile(scores_pred,
                                        100 * len(y_pred[y_pred == 1])
                                        / len(y_pred))

    # Specifies interval over which the np.linspace will be created
    x_lim_min, x_lim_max = df_floats.iloc[:, 0].min(), df_floats.iloc[:, 0].max()
    x_delta = 0.05 * (x_lim_max - x_lim_min)
    y_lim_min, y_lim_max = df_floats.iloc[:, 1].min(), df_floats.iloc[:, 1].max()
    y_delta = 0.05 * (y_lim_max - y_lim_min)
    # coordinate array for vectorized evaluation of raw anomaly scores
    # TODO: Coarser grid sometimes returns error when plotting (threshold out of
    #  [zz.min();zz.max()]).
    xx, yy = np.meshgrid(np.linspace(x_lim_min - x_delta,
                                     x_lim_max + x_delta, 100),
                         np.linspace(y_lim_min - y_delta,
                                     y_lim_max + y_delta, 100))
    # decision function calculates the raw anomaly score for every point
    zz = clf.decision_function(np.c_[xx.ravel(), yy.ravel()]) * -1
    zz = zz.reshape(xx.shape)

    # undo the scaling so the plot is in the same scale as input data
    if scaler:
        df_floats.iloc[:, [0, 1]] = scaler.inverse_transform(df_floats.iloc[:, [0, 1]])
        x_lim_min, x_lim_max = df_floats.iloc[:, 0].min(), df_floats.iloc[:, 0].max()
        x_delta = 0.05 * (x_lim_max - x_lim_min)
        y_lim_min, y_lim_max = df_floats.iloc[:, 1].min(), df_floats.iloc[:, 1].max()
        y_delta = 0.05 * (y_lim_max - y_lim_min)
        xx, yy = np.meshgrid(np.linspace(x_lim_min - x_delta, x_lim_max + x_delta, 100),
                             np.linspace(y_lim_min - y_delta, y_lim_max + y_delta, 100))

    scaler = MinMaxScaler(feature_range=(0, 1))
    threshold_scaled = scaler.fit_transform(np.array([zz.min(),
                                                      threshold,
                                                      zz.max()]).reshape(-1, 1)).item(1)

    fig = go.Figure()
    fig.add_trace(go.Contour(z=zz.ravel(), x=xx.ravel(), y=yy.ravel(),
                             colorscale=[[0, 'rgb(8,48,107)'],
                                         [threshold_scaled, 'rgb(255,255,255)'],
                                         [threshold_scaled, 'rgb(255,163,0)'],
                                         [1, 'rgb(255,163,0)']],
                             line=dict(width=0),
                             ncontours=600,

                             zmin=zz.min(), zmax=zz.max()))

    # fig1 = px.scatter(df, x='value', y='yields', color='variable')
    # for i in range(0,len(fig2.data)):
    #  fig.add_trace(fig2.data[i])

    fig2 = px.scatter(df[y_pred == 1].reset_index(), x='value', y='yields',
                      color='variable', color_discrete_sequence=['black'],
                      hover_name='created_on')

    fig3 = px.scatter(df[y_pred == 0].reset_index(), x='value', y='yields',
                      color='variable', color_discrete_sequence=['white'],
                      hover_name='created_on')
    fig3.update_traces(marker=dict(line=dict(width=0.2)),
                       selector=dict(mode='markers'))
    for i in range(0, len(fig3.data)):
        fig.add_trace(fig3.data[i])
    for i in range(0, len(fig2.data)):
        fig.add_trace(fig2.data[i])

    fig.add_trace(go.Histogram(
        y=df_floats.iloc[:, 1].values,
        xaxis='x2',
        showlegend=False,
        marker=dict(
            color='rgba(0,0,0,1)'
        )
    ))
    fig.add_trace(go.Histogram(
        x=df_floats.iloc[:, 0].values,
        yaxis='y2',
        showlegend=False,
        marker=dict(
            color='rgba(0,0,0,1)'
        )
    ))

    fig.update_layout(
        autosize=True,
        xaxis=dict(
            zeroline=False,
            domain=[0, 0.85],
            showgrid=False,
            title=df_floats.columns[0]
        ),
        yaxis=dict(
            zeroline=False,
            domain=[0, 0.85],
            showgrid=False,
            title=df_floats.columns[1]
        ),
        xaxis2=dict(
            zeroline=False,
            domain=[0.85, 1],
            showgrid=False
        ),
        yaxis2=dict(
            zeroline=False,
            domain=[0.85, 1],
            showgrid=False
        ),
        height=800,
        width=1000,
        bargap=0,
        hovermode='closest',
        showlegend=True,
        legend_x=1.2,
        title='Heatmap of outlier score classified by ' + clf_name
    )

    plotly.offline.plot(fig)

    return


def plot_outlier_detection(
        df_floats: pd.DataFrame,
        y_labels: np.ndarray,
        clf,
        clf_name: str = None,
        scaler=None,
):
    """Return contourf plot of the anomaly detection model.

    Plots contourf plot of decision scores marking area
    where observations would be considered inliers.

    Parameters
        ----------
        df_floats:
            pd.DataFrame with floats as the elements.
        y_labels:
            numpy array of the same length as df_floats that assigns
            0/1 (inlier/outlier) to each observation according to fitted model.
        clf:
            fitted model.
        clf_name:
            name of fitted model.
        scaler:
            estimator that was used to change range of df_floats DataFrame

    Returns
        -------
        Contourf plot.
    """
    if df_floats.shape[1] > 2:
        print("Plotting first two variables...")
    elif df_floats.shape[1] < 2:
        print("Sorry can not plot less than two variables.")
        return

    # predict raw anomaly score
    y_scores = clf.decision_function(df_floats.iloc[:, [0, 1]]) * -1
    # threshold value to consider a datapoint inlier or outlier
    threshold = stats.scoreatpercentile(
        y_scores, 100 * len(y_labels[y_labels == 1]) / len(y_labels)
    )

    # Specifies interval over which the np.linspace will be created
    x_lim_min, x_lim_max = df_floats.iloc[:, 0].min(), df_floats.iloc[:, 0].max()
    x_delta = 0.05 * (x_lim_max - x_lim_min)
    y_lim_min, y_lim_max = df_floats.iloc[:, 1].min(), df_floats.iloc[:, 1].max()
    y_delta = 0.05 * (y_lim_max - y_lim_min)
    # coordinate array for vectorized evaluation of raw anomaly scores
    xx, yy = np.meshgrid(
        np.linspace(x_lim_min - x_delta, x_lim_max + x_delta, 100),
        np.linspace(y_lim_min - y_delta, y_lim_max + y_delta, 100),
    )
    # decision function calculates the raw anomaly score for every point
    zz = clf.decision_function(np.c_[xx.ravel(), yy.ravel()]) * -1
    zz = zz.reshape(xx.shape)

    # undo the scaling so the plot is in the same scale as input data
    if scaler:
        df_floats.iloc[:, [0, 1]] = scaler.inverse_transform(df_floats.iloc[:, [0, 1]])
        x_lim_min, x_lim_max = df_floats.iloc[:, 0].min(), df_floats.iloc[:, 0].max()
        x_delta = 0.05 * (x_lim_max - x_lim_min)
        y_lim_min, y_lim_max = df_floats.iloc[:, 1].min(), df_floats.iloc[:, 1].max()
        y_delta = 0.05 * (y_lim_max - y_lim_min)
        xx, yy = np.meshgrid(
            np.linspace(x_lim_min - x_delta, x_lim_max + x_delta, 100),
            np.linspace(y_lim_min - y_delta, y_lim_max + y_delta, 100),
        )

    # inliers_1 - inlier feature 1,  inliers_2 - inlier feature 2
    inliers_1 = (df_floats.iloc[:, 0][y_labels == 0]).values.reshape(-1, 1)
    inliers_2 = (df_floats.iloc[:, 1][y_labels == 0]).values.reshape(-1, 1)
    # outliers_1 - outlier feature 1, outliers_2 - outlier feature 2
    outliers_1 = df_floats.iloc[:, 0][y_labels == 1].values.reshape(-1, 1)
    outliers_2 = df_floats.iloc[:, 1][y_labels == 1].values.reshape(-1, 1)

    plt.figure(figsize=(10, 10))
    # fill blue map colormap from minimum anomaly score to threshold value
    plt.contourf(xx, yy, zz, levels=np.linspace(zz.min(), threshold, 50),
                 cmap=plt.cm.Blues_r)
    plt.colorbar(
        plt.contourf(xx, yy, zz, levels=np.linspace(zz.min(), threshold, 50),
                     cmap=plt.cm.Blues_r)
    )
    # fill orange contour lines where range of anomaly score is
    # from threshold to maximum anomaly score
    plt.contourf(xx, yy, zz, levels=[threshold, zz.max()], colors="orange")
    # draw red contour line where anomaly score is equal to threshold
    a = plt.contour(xx, yy, zz, levels=[threshold], linewidths=2, colors="red")
    # draw inliers as white dots
    b = plt.scatter(inliers_1, inliers_2, c="white", s=20, edgecolor="k")
    # draw outliers as black dots
    c = plt.scatter(outliers_1, outliers_2, c="black", s=20, edgecolor="k")

    plt.axis("tight")
    # loc=2 is used for the top left corner
    plt.legend(
        [a.collections[0], b, c],
        ["learned decision function", "inliers", "outliers"],
        prop=matplotlib.font_manager.FontProperties(size=20),
        loc=2,
    )

    plt.xlim((x_lim_min - x_delta, x_lim_max + x_delta))
    plt.ylim((y_lim_min - y_delta, y_lim_max + y_delta))
    plt.xlabel(df_floats.columns[0])
    plt.ylabel(df_floats.columns[1])
    if clf_name:
        plt.title(clf_name)
    plt.show()

    return
